(*
  Copyright (C) BitBlaze, 2009-2010. 

#Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

#THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*)
(*
*  vine_alphavary.ml
 *)
(**
   Code for alpha-varying and descoping programs.
*)


(** class used for renaming variables for both alpha-varying and
    descoping *)
class var_renamer :
  object
    val ctx : (int, Vine.var) Hashtbl.t
    method alpha_vary_decl : Vine.decl -> Vine.var
    method alpha_vary_decls : Vine.decl list -> Vine.decl list
    method exit_decl : Vine.decl -> unit
    method exit_decls : Vine.decl list -> unit
    method init_ctx : Vine.decl list -> unit
    method ctx_to_lst : unit -> Vine.decl list 
    method visit_alvalue : Vine.lvalue -> Vine.lvalue Vine.visit_action
    method visit_binding :
      Vine.lvalue * Vine.exp -> (Vine.lvalue * Vine.exp) Vine.visit_action
    method visit_decl : Vine.decl -> Vine.decl Vine.visit_action
    method visit_exp : Vine.exp -> Vine.exp Vine.visit_action
    method visit_rlvalue : Vine.lvalue -> Vine.lvalue Vine.visit_action
    method visit_stmt : Vine.stmt -> Vine.stmt Vine.visit_action
  end



(** alpha-vary a single expression *)
val alpha_vary_exp : Vine.exp -> Vine.exp

(** alpha-vary a single statement *)
val alpha_vary_stmt : Vine.stmt -> Vine.stmt

(** alpha-vary a sequence of statements. *)
val alpha_vary_stmts : Vine.stmt list -> Vine.stmt list

(** alpha-vary a complete program, except for globals *)
val alpha_vary_program : Vine.program -> Vine.program

(** remove all scoping from a program. Note variable names will be
    globally unique. 
*)
val descope_program : Vine.program -> Vine.program

(** descope a single function. *)
val descope_function : Vine.stmt -> Vine.stmt

val descope_stmts : var_renamer -> Vine.stmt list -> Vine.program 
