(*
  Copyright (C) BitBlaze, 2009-2010. 

#Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

#THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*)
type 'a loop
type combine_type = KeepDisjoint | UniqueHeader

module LOOP :functor (Lang : sig type t end) ->
sig
  val make_reducible : 
    (Lang.t Vine_cfg.cfg -> 
      (Lang.t Vine_cfg.bb, Lang.t Vine_cfg.bb) Hashtbl.t -> unit) -> 
    Lang.t Vine_cfg.cfg -> unit

  val get_reducible : 
    (Lang.t Vine_cfg.cfg -> 
      (Lang.t Vine_cfg.bb, Lang.t Vine_cfg.bb) Hashtbl.t -> unit) -> 
    (Lang.t Vine_cfg.cfg as 'b) -> 'b

  val get_loops : ?combine_method:combine_type -> Lang.t Vine_cfg.cfg -> 
    Lang.t loop list
    
  val get_loops_from_bb : ?combine_method:combine_type -> 
    Lang.t Vine_cfg.cfg -> Lang.t Vine_cfg.bb -> Lang.t loop list
    
  val unroll_loop_by_backedges :
    ((Lang.t Vine_cfg.cfg as 'a) -> ((Lang.t Vine_cfg.bb as 'b), 'b) Hashtbl.t
      -> unit) ->
    ('a -> ('b * 'b) list -> ('b, 'b) Hashtbl.t -> unit) ->
    'a -> 'b list -> ('b * 'b) list -> int -> unit
    
  val unroll_loop :
    ((Lang.t Vine_cfg.cfg as 'a) -> ((Lang.t Vine_cfg.bb as 'b), 'b) Hashtbl.t
      -> unit) ->
    ('a -> ('b * 'b) list -> ('b, 'b) Hashtbl.t -> unit) ->
    'a -> Lang.t loop -> int -> unit
    
end

val get_loop_header : 'a loop -> 'a Vine_cfg.bb
  
val get_loop_body : 'a loop -> 'a Vine_cfg.bb list
    
val get_loop_back_nodes : 'a loop -> 'a Vine_cfg.bb list
  
val get_loop_exit_edges : 'a loop -> 'a #Vine_cfg.cfg ->
  ('a Vine_cfg.bb * 'a Vine_cfg.bb) list
    
val get_inner_loops : 'a loop -> 'a #Vine_cfg.cfg -> 'a loop list
  
val get_outer_loops : 'a loop -> 'a #Vine_cfg.cfg -> 'a loop list
  
val pp_loop : (string -> unit) -> 'a loop -> unit
  
val make_reducible : Vine.stmt list Vine_cfg.cfg -> unit
  
val get_reducible : (Vine.stmt list Vine_cfg.cfg as 'b) -> 'b
  
val get_loops : ?combine_method:combine_type -> 
  Vine.stmt list Vine_cfg.cfg -> Vine.stmt list loop list
  
val get_loops_from_bb : ?combine_method:combine_type -> 
  Vine.stmt list Vine_cfg.cfg -> 
  Vine.stmt list Vine_cfg.bb -> Vine.stmt list loop list
    
val unroll_loop_by_backedges :
  ?fwd_jmps:(Vine.stmt list Vine_cfg.cfg -> 
    Vine_cfg.label -> Vine_cfg.label * Vine.stmt list Vine_cfg.bb) ->
  Vine.stmt list Vine_cfg.cfg -> 
  Vine.stmt list Vine_cfg.bb list -> 
  (Vine.stmt list Vine_cfg.bb * Vine.stmt list Vine_cfg.bb) list ->
    int ->
  unit
      
val unroll_loop :
    ?fwd_jmps:(Vine.stmt list Vine_cfg.cfg -> 
      Vine_cfg.label -> Vine_cfg.label * Vine.stmt list Vine_cfg.bb) ->
  Vine.stmt list Vine_cfg.cfg -> Vine.stmt list loop -> int -> unit
  
