(*
  Copyright (C) BitBlaze, 2009-2010. 

#Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

#THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*)

class virtual concrete_memory : object
  method virtual store_byte : Int64.t -> int -> unit
  method virtual maybe_load_byte : Int64.t -> int option
  method virtual clear : unit -> unit

  method measure_size : int

  method load_byte : int64 -> int 

  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit

  method store_page : int64 -> string -> unit

  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
    
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option
end

class concrete_string_memory : object
  inherit concrete_memory

  method store_byte : Int64.t -> int -> unit
  method maybe_load_byte : Int64.t -> int option
  method clear : unit -> unit
  method measure_size : int
  method load_byte : int64 -> int 
  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit
  method store_page : int64 -> string -> unit
  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option
end

class concrete_hash_memory : object
  inherit concrete_memory

  method store_byte : Int64.t -> int -> unit
  method maybe_load_byte : Int64.t -> int option
  method clear : unit -> unit
  method measure_size : int
  method load_byte : int64 -> int 
  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit
  method store_page : int64 -> string -> unit
  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option
end

class virtual concrete_memory_w_reset : object
  inherit concrete_memory

  method virtual store_byte : Int64.t -> int -> unit
  method virtual maybe_load_byte : Int64.t -> int option
  method virtual clear : unit -> unit
  method measure_size : int
  method load_byte : int64 -> int 
  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit
  method store_page : int64 -> string -> unit
  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option

  method virtual make_snap : unit -> unit
  method virtual reset : unit -> unit
end

class concrete_snapshot_memory :
  concrete_memory_w_reset -> concrete_memory_w_reset -> object
  inherit concrete_memory
    
  method store_byte : Int64.t -> int -> unit
  method maybe_load_byte : Int64.t -> int option
  method clear : unit -> unit
  method measure_size : int
  method load_byte : int64 -> int 
  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit
  method store_page : int64 -> string -> unit
  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option

  method make_snap : unit -> unit
  method reset : unit -> unit
end

class parallel_check_memory :
  concrete_memory_w_reset -> concrete_memory_w_reset -> object
  inherit concrete_memory

  method store_byte : Int64.t -> int -> unit
  method maybe_load_byte : Int64.t -> int option
  method clear : unit -> unit
  method measure_size : int
  method load_byte : int64 -> int 
  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit
  method store_page : int64 -> string -> unit
  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option

  method make_snap : unit -> unit
  method reset : unit -> unit
end

class string_maybe_memory : object
  inherit concrete_memory

  method store_byte : Int64.t -> int -> unit
  method maybe_load_byte : Int64.t -> int option
  method clear : unit -> unit
  method measure_size : int
  method load_byte : int64 -> int 
  method store_short : int64 -> int   -> unit
  method store_word  : int64 -> int64 -> unit
  method store_long  : int64 -> int64 -> unit
  method store_page : int64 -> string -> unit
  method load_short : int64 -> int
  method load_word  : int64 -> int64
  method load_long  : int64 -> int64
  method maybe_load_short : int64 -> int   option
  method maybe_load_word  : int64 -> int64 option
  method maybe_load_long  : int64 -> int64 option
end
